# -------------------------------------------------------------------------------------------------
# test_math_mp_xi2.py: tests Calculating using Pool of Workers in Multiprocessing                              
#
# Theory:      Multiprocessing
#
# Description: Multiprocessing should be faster than threading
#              the current threading system in Python does not support multiprocssing
#              the speed fo multtiprocessing depends (among other things) on the number of cores embedded inside CPU
#
# class multiprocessing.Process([group[, target[, name[, args[, kwargs]]]]]) 
# 
#    Process objects represent activity that is run in a separate process. 
#    The Process class has equivalents of all the methods of threading.Thread.
#    
#constructor
#    should always be called with keyword arguments. 
#    
#group
#    should always be None; it exists solely for compatibility with Thread. 
#    
#target
#    is the callable object to be invoked by the run() method. 
#    It defaults to None, meaning nothing is called. 
#    
#name
#    is the process name
#    by default, a unique name is constructed of the form Process-N1:N2:...:Nk
#    where N1,N2,...,Nk is a sequence of integers whose length is determined by the generation of the process. 
#    
#args
#    is the argument tuple for the target invocation
#    
#kwargs
#    is a dictionary of keyword arguments for the target invocation. By default, no arguments are passed to target.
#    
#If a subclass overrides the constructor, it must make sure it invokes the base class constructor (Process.__init__()) before doing anything else to the process.
#    
#run() 
#    Method representing the process activity.
#    You may override this method in a subclass. 
#    The standard run() method invokes the callable object passed to the object constructor as the target argument, if any, 
#    with sequential and keyword arguments taken from the args and kwargs arguments, respectively.
#    
#start() 
#    Start the process activity.
#    This must be called at most once per process object. 
#    It arranges for the object run() method to be invoked in a separate process.
#    
#join([timeout]) 
#    Block the calling thread until the process whose join() method is called terminates or until the optional timeout occurs.
#    If timeout is None then there is no timeout.
#    A process can be joined many times.
#    A process cannot join itself because this would cause a deadlock. It is an error to attempt to join a process before it has been started.
#    
#name 
#    Return the process name.
#    
#Process.name = name 
#    Set the process name.
#    The name is a string used for identification purposes only. 
#    It has no semantics. 
#    Multiple processes may be given the same name. 
#    The initial name is set by the constructor.
#    
#is_alive() <--- OK
#    Return whether the process is alive.
#    Roughly, a process object is alive from the moment the start() method returns until the child process terminates.
#    
#daemon 
#    Return the process daemon flag., this is a boolean.
#    
#Process.daemon = daemonic 
#    Set the process daemon flag to the Boolean value daemonic. 
#    This must be called before start() is called.
#    The initial value is inherited from the creating process.
#    When a process exits, it attempts to terminate all of its daemonic child processes.
#    Note that a daemonic process is not allowed to create child processes. 
#    Otherwise a daemonic process would leave its children orphaned if it gets terminated when its parent process exits.
#        
#In addition process objects also support the following methods:
#    
#pid 
#    Return the process ID. Before the process is spawned, this will be None.
#    
#exitcode 
#    Return the child exit code. This will be None if the process has not yet terminated. A negative value -N indicates that the child was terminated by signal N.
#
#authkey 
#    Return the process authentication key (a byte string).
#    When multiprocessing is initialized the main process is assigned a random string using os.random().
#    When a Process object is created, it will inherit the authentication key of its parent process, although this may be changed using Process.authkey below.
#    See Authentication keys.
# 
#Process.authkey = authkey 
# Set the process authentication key which must be a byte string.
# 
#terminate() 
#    Terminate the process. 
#    On Unix this is done using the SIGTERM signal; 
#    on Windows TerminateProcess is used. Note that exit handlers and finally clauses, etc., will not be executed.
#    Note that descendant processes of the process will not be terminated, they will simply become orphaned
#    Warning
#    If this method is used when the associated process is using a pipe or queue 
#    then the pipe or queue is liable to become corrupted and may become unusable by other process. 
#    Similarly, if the process has acquired a lock or semaphore etc. then terminating it is liable to cause other processes to deadlock.
#    Note that the start(), join(), is_alive() and get_exit_code() methods should only be called by the process that created the process object.
# -------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------
# Imports from standard modules                              
# -------------------------------------------------------------------------------------------------
import os
import sys
from ctypes import Structure, c_double
import time
import threading
import gc
import random
import queue # <--- NOT Queue as in the original example
import operator

# -------------------------------------------------------------------------------------------------
# Imports from multiprocessing package                              
# -------------------------------------------------------------------------------------------------
import multiprocessing
from multiprocessing import Process
from multiprocessing import Queue
from multiprocessing import Pipe
from multiprocessing import Lock
from multiprocessing import Manager
from multiprocessing import Pool
from multiprocessing import current_process
from multiprocessing import freeze_support
from multiprocessing.sharedctypes import Value, Array
from multiprocessing.managers import BaseManager, BaseProxy


# -------------------------------------------------------------------------------------------------
# yourDragonXi imports                               
# -------------------------------------------------------------------------------------------------
from process2xi import Process2Xi
from queue2xi  import Queue2Xi
from pipe2xi import Pipe2Xi
from lock2xi import Lock2Xi
from rlock2xi import RLock2Xi
from manager2xi import Manager2Xi 
from pool2xi import Pool2Xi
from value2xi import Value2Xi
from arrayfunction2xi import ArrayFunction2Xi #<--- note: this must be called (Array2Xi uses the class from managers.py)
from basemanager2xi import BaseManager2Xi
from baseproxy2xi import BaseProxy2Xi
from condition2xi import Condition2Xi  




# -------------------------------------------------------------------------------------------------
# Function run by worker processes
# -------------------------------------------------------------------------------------------------

def worker(input, output):
    for func, args in iter(input.get, 'STOP'):
        result = calculate(func, args)
        output.put(result)

# -------------------------------------------------------------------------------------------------
# Function used to calculate result
# -------------------------------------------------------------------------------------------------

def calculate(func, args):
    result = func(*args)
    return '%s says that %s%s = %s' % \
        (current_process().name, func.__name__, args, result)

# -------------------------------------------------------------------------------------------------
# Functions referenced by tasks
# -------------------------------------------------------------------------------------------------

def mul(a, b):
    time.sleep(0.5*random.random())
    return a * b

def plus(a, b):
    time.sleep(0.5*random.random())
    return a + b

# -------------------------------------------------------------------------------------------------
# Function test1: tests calculations with standard multiprocessing
# -------------------------------------------------------------------------------------------------
def test1():
    print("\nFunction test1: tests calculations with standard multiprocessing")
    NUMBER_OF_PROCESSES = 4
    TASKS1 = [(mul, (i, 7)) for i in range(20)]
    TASKS2 = [(plus, (i, 8)) for i in range(10)]

    # Create queues
    task_queue = Queue()
    done_queue = Queue()

    # Submit tasks
    for task in TASKS1:
        task_queue.put(task)

    # Start worker processes
    for i in range(NUMBER_OF_PROCESSES):
        Process(target=worker, args=(task_queue, done_queue)).start()

    # Get and print results
    print ('Unordered results:')
    for i in range(len(TASKS1)):
        print ('\t', done_queue.get())

    # Add more tasks using `put()`
    for task in TASKS2:
        task_queue.put(task)

    # Get and print some more results
    for i in range(len(TASKS2)):
        print ('\t', done_queue.get())

    # Tell child processes to stop
    for i in range(NUMBER_OF_PROCESSES):
        task_queue.put('STOP')

# -------------------------------------------------------------------------------------------------
# Function test2: tests calculations with dragonized multiprocessing
# -------------------------------------------------------------------------------------------------
def test2():
    print("\nFunction test2: tests calculations with dragonized multiprocessing")
    NUMBER_OF_PROCESSES = 4
    TASKS1 = [(mul, (i, 7)) for i in range(20)]
    TASKS2 = [(plus, (i, 8)) for i in range(10)]

    # Create queues
    task_queue = Queue2Xi()
    done_queue = Queue2Xi()

    # Submit tasks
    for task in TASKS1:
        task_queue.put(task)

    # Start worker processes
    for i in range(NUMBER_OF_PROCESSES):
        Process2Xi(target=worker, args=(task_queue, done_queue)).start()

    # Get and print results
    print ('Unordered results:')
    for i in range(len(TASKS1)):
        print ('\t', done_queue.get())

    # Add more tasks using `put()`
    for task in TASKS2:
        task_queue.put(task)

    # Get and print some more results
    for i in range(len(TASKS2)):
        print ('\t', done_queue.get())

    # Tell child processes to stop
    for i in range(NUMBER_OF_PROCESSES):
        task_queue.put('STOP')        
        
# -------------------------------------------------------------------------------------------------
#                               
# -------------------------------------------------------------------------------------------------
if __name__ == '__main__':
    multiprocessing.freeze_support()
    test1()
    test2()
# -------------------------------------------------------------------------------------------------
# test2: Example Output using dragonized multiprocessing                                
# -------------------------------------------------------------------------------------------------
# Unordered results:
# 	 Process2Xi-1 says that mul(0, 7) = 0
# 	 Process2Xi-4 says that mul(3, 7) = 21
# 	 Process2Xi-4 says that mul(5, 7) = 35
# 	 Process2Xi-2 says that mul(1, 7) = 7
# 	 Process2Xi-1 says that mul(2, 7) = 14
# 	 Process2Xi-4 says that mul(6, 7) = 42
# 	 Process2Xi-3 says that mul(4, 7) = 28
# 	 Process2Xi-4 says that mul(9, 7) = 63
# 	 Process2Xi-3 says that mul(10, 7) = 70
# 	 Process2Xi-1 says that mul(8, 7) = 56
# 	 Process2Xi-1 says that mul(13, 7) = 91
# 	 Process2Xi-2 says that mul(7, 7) = 49
# 	 Process2Xi-1 says that mul(14, 7) = 98
# 	 Process2Xi-1 says that mul(16, 7) = 112
# 	 Process2Xi-4 says that mul(11, 7) = 77
# 	 Process2Xi-3 says that mul(12, 7) = 84
# 	 Process2Xi-3 says that mul(19, 7) = 133
# 	 Process2Xi-2 says that mul(15, 7) = 105
# 	 Process2Xi-1 says that mul(17, 7) = 119
# 	 Process2Xi-4 says that mul(18, 7) = 126
# 	 Process2Xi-4 says that plus(3, 8) = 11
# 	 Process2Xi-4 says that plus(4, 8) = 12
# 	 Process2Xi-2 says that plus(0, 8) = 8
# 	 Process2Xi-3 says that plus(1, 8) = 9
# 	 Process2Xi-1 says that plus(2, 8) = 10
# 	 Process2Xi-1 says that plus(7, 8) = 15
# 	 Process2Xi-4 says that plus(5, 8) = 13
# 	 Process2Xi-2 says that plus(6, 8) = 14
# 	 Process2Xi-1 says that plus(9, 8) = 17
# 	 Process2Xi-3 says that plus(8, 8) = 16 
# 
# -------------------------------------------------------------------------------------------------
# test1: Example Output using standard multiprocessing                                
# -------------------------------------------------------------------------------------------------
# 
# Unordered results:
# 	 Process-2 says that mul(3, 7) = 21
# 	 Process-3 says that mul(2, 7) = 14
# 	 Process-4 says that mul(1, 7) = 7
# 	 Process-4 says that mul(6, 7) = 42
# 	 Process-2 says that mul(4, 7) = 28
# 	 Process-3 says that mul(5, 7) = 35
# 	 Process-1 says that mul(0, 7) = 0
# 	 Process-3 says that mul(9, 7) = 63
# 	 Process-2 says that mul(8, 7) = 56
# 	 Process-4 says that mul(7, 7) = 49
# 	 Process-1 says that mul(10, 7) = 70
# 	 Process-2 says that mul(12, 7) = 84
# 	 Process-3 says that mul(11, 7) = 77
# 	 Process-4 says that mul(13, 7) = 91
# 	 Process-2 says that mul(16, 7) = 112
# 	 Process-1 says that mul(14, 7) = 98
# 	 Process-3 says that mul(15, 7) = 105
# 	 Process-2 says that mul(18, 7) = 126
# 	 Process-4 says that mul(17, 7) = 119
# 	 Process-1 says that mul(19, 7) = 133
# 	 Process-2 says that plus(1, 8) = 9
# 	 Process-3 says that plus(0, 8) = 8
# 	 Process-3 says that plus(5, 8) = 13
# 	 Process-1 says that plus(3, 8) = 11
# 	 Process-1 says that plus(7, 8) = 15
# 	 Process-3 says that plus(6, 8) = 14
# 	 Process-3 says that plus(9, 8) = 17
# 	 Process-1 says that plus(8, 8) = 16
# 	 Process-4 says that plus(2, 8) = 10
# 	 Process-2 says that plus(4, 8) = 12
# 
# -------------------------------------------------------------------------------------------------
# End Of test_math_mp_xi2.py                               
# -------------------------------------------------------------------------------------------------